#include <stdio.h>
#include "MP4moviewrapper_enc.h" 

#ifdef BSAC
extern int codec_type;
extern int bsac_sf;
extern int bsac_ch;
#endif
/* decoder-related functions */
MP4Err MP4InitMovie(MP4MovieStruct* movie, const char* MovieFileName, unsigned int *corepresent,
						   unsigned int* samplecount, unsigned int* bitrate, unsigned int* totallayers)

{
	u32 i, trackcount, firsttime=1;
	MP4Err err=0;

#ifdef SAMPLEGROUP_ELST
  /* init values from movie handle */
  for(i=0; i<NUM_LAYERS; i++){
    movie->sampleDescriptionIndex[i] = 0;
    movie->samplesPerFrame[i]        = 0;

    movie->sampleBufH_size[i]        = 0;
    movie->preRoll[i]                = 0;
    movie->eof[i]                    = 0;
    movie->outSamples[i]             = 0;
    movie->firstFrame[i]             = 1;
  }
#endif

	if(err = MP4OpenMovieFile( &(movie->moov), MovieFileName, MP4OpenMovieNormal)) goto bail;
	if ((err = MP4GetMovieTrackCount(movie->moov, &trackcount))) goto bail;
 
	for(i=0;i<trackcount;i++)
	{
	  unsigned int ObjectType, StreamType, BufferSize, MaxBitrate,streamPriority,
					AvgBitrate, AACBitrate=0;
		unsigned int audioObjectType, audioObjectTypeExt, sampIndex, channelConfig, 
					LLEmainstream, reservedBit, pcmLength, frameIndex, samplingfreq;
	
		/* I2R: changed to unsigned type (28 May 07) */
		unsigned char* ptrH;

		if(err = MP4GetMovieIndTrack( movie->moov, i+1, &(movie->trak[i]) )) break;
		if(err = MP4CreateTrackReader( movie->trak[i], &(movie->trackReader[i]) )) goto bail;
		if(err = MP4GetTrackMedia( movie->trak[i], &(movie->media[i]) ))goto bail;
		if(err = MP4NewHandle( 0, &(movie->decoderSpecificInfoH[i])))goto bail;

		/* I2R: included stream priority parameter (28 May 07) */
		if (err = MP4GetMediaDecoderInformation( movie->media[i],
                                                         1,
                                                         &ObjectType,
                                                         &StreamType, 
                                                         &BufferSize,
                                                         NULL,
                                                         &MaxBitrate,
                                                         &AvgBitrate, 
                                                         movie->decoderSpecificInfoH[i]
                                                         )) goto bail;

		bitrate[i]=AvgBitrate/1000;
		
		if ( ObjectType != 0x40 ) {err = MP4BadDataErr; goto bail; }
		if ( StreamType != 0x05 ) {err = MP4BadDataErr; goto bail; }
				
		ptrH = *(movie->decoderSpecificInfoH[i]);
		
		audioObjectType	= (((*ptrH)>>3)&0x1f);
		if(audioObjectType==31)
		{
			audioObjectTypeExt	= (((*ptrH)&0x07)<<3) | (((*(ptrH+1))>>5)&0x07);
			audioObjectType	= 32+audioObjectTypeExt;
		}

		if((trackcount==1) && (audioObjectType!=38) )
		{
#if 0
			channelConfig=2;
			*corepresent=1;
			reservedBit=0;
			pcmLength=0;	
			if(err = MP4GetMediaTimeScale(movie->media[i], &samplingfreq))  goto bail;
			if(err = MP4GetMediaSampleCount(movie->media[i], samplecount))  goto bail;
			
			switch(samplingfreq)
			{
				case 11025:	sampIndex = 10; break;
				case 22050:	sampIndex = 7; break;
				case 44100:	sampIndex = 4; break;
				case 48000:	sampIndex = 3; break;
				default: printf("sampling rate currently not supported!\n"); return 0;
			}
#else
                        printf("audio object type %d in single layer mode not supported!\n", audioObjectType);
                        goto bail;
#endif
		}

		if(firsttime && ((audioObjectType==37)||(audioObjectType==38)) )
		{
			sampIndex		= ((*(ptrH+1))>>1)&0x0f;	samplingfreq	= sampling_rate[sampIndex];			
			
			/* I2R: modified to include sampIndex and samplingFreq = 192kHz (21 Jun 07) */
			if(sampIndex<0xf)
			{
                          /*    | AOT   | AOText | s_idx | cc   | PCM   | core  | lle   | res   | frame |  */
                          /*    0       5        11      15     19      22      23      24      25      28 */
	
                          channelConfig	= ((*(ptrH+2))>>5)&0x0f;	/* channel configuration */
                          pcmLength     = ((*(ptrH+2))>>2)&0x03;
                          *corepresent	= ((*(ptrH+2))>>1)&0x01;
                          LLEmainstream	= ((*(ptrH+2))>>0)&0x01;
                          reservedBit	= ((*(ptrH+3))>>7)&0x01;
                          frameIndex	= ((*(ptrH+3))>>4)&0x07;
			}
			else
			{
                          /*    | AOT   | AOText | s_idx | sr   | cc    | PCM   | core  | lle   | res   | frame | */
                          /*    0       5        11      15     39      43      46      47      48      49      52 */
	
                          samplingfreq	= 192000;		/* todo: we should read the samplingfreq from the ASC */
                          channelConfig	= ((*(ptrH+4))&0x01)<<3 | ((*(ptrH+5))>>5)&0x07;
                          pcmLength	= ((*(ptrH+5))>>2)&0x03;
                          *corepresent	= ((*(ptrH+5))>>1)&0x01;
                          LLEmainstream	= ((*(ptrH+5))>>0)&0x01;
                          reservedBit	= ((*(ptrH+6))>>7)&0x01;
                          frameIndex	= ((*(ptrH+6))>>4)&0x07;
			}
			if(pcmLength>3){err = MP4BadDataErr; goto bail; }
			if(!LLEmainstream){err = MP4BadDataErr; goto bail; }
			if(frameIndex>2) {err = MP4BadDataErr; goto bail; }
                        
			if(err = MP4GetMediaSampleCount(movie->media[i], samplecount))  goto bail;
			firsttime=0;			
		}
#ifdef BSAC
		if(firsttime && ((audioObjectType==22)) )
		{
                  sampIndex	= (((*ptrH)&0x07)<<1) | (((*(ptrH+1))>>7)&0x01);
                  samplingfreq	= sampling_rate[sampIndex];			
                  channelConfig	= (((*(ptrH+1))>>3)&0x0f);	/* channel configuration */
                  codec_type=1;
                  bsac_sf = (int)samplingfreq;
                  bsac_ch = channelConfig;
		}
#endif
	}
	*totallayers=i;

	for (i=0; i<*totallayers; i++)
          if(err = MP4NewHandle( 1024*36, &(movie->sampleBufH[i]) )) goto bail; /* I2R (27 Jun 07) */

        /* check track references */
        for (i=0; i<*totallayers; i++) {
          if (i > 0) {

            MP4Track trak;

            err = MP4GetTrackReference( movie->trak[i],
                                        MP4StreamDependencyReferenceType,
                                        1,
                                        &trak
                                        );

            if ( (err < 0) || (trak != movie->trak[i-1]) ) {
              printf("no valid track reference found for track %d!\n", i);
              err = 0;
            } else {
              /* printf("track reference found (%d -> %d)\n", i, i-1); */
            }
          }
        }

bail:	
	return err;
}



#ifndef SAMPLEGROUP_ELST
MP4Err MP4GetNextAUfromMovie(MP4MovieStruct* movie, unsigned int *sampleBufLen,
                                                        unsigned int* duration, unsigned int numlayers  )
{
  unsigned int i, unitSize, sampleFlags, cts, dts;

        MP4Err err=0;

  for (i=0;i<numlayers;i++)
  {
    if(err = MP4TrackReaderGetNextAccessUnitWithDuration( movie->trackReader[i],
                        movie->sampleBufH[i],&sampleBufLen[i], &sampleFlags,
                        &cts, &dts,duration )) goto bail;
  }
bail:	
  return err;
}

#else
MP4Err MP4GetNextAUfromMovie(MP4MovieStruct* movie, 
                             unsigned char sampleBuf[][1<<16], 
                             unsigned int *sampleBufLen,
                             int* outSamples, 
                             unsigned int numlayers  )
{
  unsigned int i, unitSize, sampleFlags, cts, dts;
  int tmp_duration;
  
  MP4Err err=0;

  /* set outsamples to 0 */
  *outSamples = 0;
  
  for (i=0;i<numlayers;i++)
  {	
    
    /* copy delay samples from pre-roll to output */
    if(movie->preRoll[i] == -1){
      if(movie->sampleBufH_size[i] <= (1<<16)){
        memcpy(sampleBuf[i], *movie->sampleBufH[i], movie->sampleBufH_size[i]);
        sampleBufLen[i] = movie->sampleBufH_size[i];
        *outSamples = movie->outSamples[i];
      } else {
        return 1;
      }
    }
    
    if(err = MP4TrackReaderGetNextAccessUnitWithDuration( movie->trackReader[i],
      movie->sampleBufH[i],&movie->sampleBufH_size[i], &sampleFlags,
      &cts, &dts, &tmp_duration )){
      
      /* return in case of any non-eof error*/
      if (err != MP4EOF) {      
        goto bail;		
      }
      
      /* here we have EOF from libisomedia for this track.
      in case of mp4ff and samplegroups (-> preroll=-1), we hold back
      the EOF error if it occurs for the first time (m_eof[trackNum] == 0)
      */
      
      if((movie->preRoll[i] == -1) && (movie->eof[i] == 0)){
        movie->eof[i] = 1;
        err = ISONoErr;
      }
      
      /* return eof; we asume that file ends if one track ends */
      if(err != ISONoErr){
        return err;
      }
    }
    
    /* calculate number of samples to playout */
    /* asume track last track as sls and responsible for sample output */
    if (dts < 0) {
      /* at the beginnin of file */
      movie->outSamples[i] = dts;
    } else {
      movie->outSamples[i] = tmp_duration; 
    }
    
    if (movie->firstFrame[i] == 1) {
      
      int sampleNumber =  -1;
      int groupIndex   =  -1;
      unsigned int group_type = 'r' << 24 | 'o' << 16 | 'l' << 8 | 'l';

      /* reset first frame flag */
      movie->firstFrame[i] = 0;
      
      /* let's see if we need preroll handling */
      err = MP4TrackReaderGetCurrentSampleNumber(movie->trackReader[i], &sampleNumber);
      if (err) return err;
      
      ISOGetSampletoGroupMap(movie->media[i], group_type, sampleNumber, &groupIndex);
      
      /* if we can not retrieve SampleGroup Information we continue normal decoding */ 
      if (groupIndex > 0) {
        /* sample belongs to a group defining a preRoll */
        
        unsigned char* desc_buff = NULL;
        ISOHandle description = NULL;
        err = ISONewHandle( 2, &description); 
        
        err = ISOGetGroupDescription(movie->media[i], group_type, groupIndex, description);
        desc_buff = (unsigned char *) *description;
        movie->preRoll[i] = desc_buff[0] << 8 | desc_buff[1];
        
        /* right now we only allow preRoll of 0 and -1 */
        if((movie->preRoll[i] != -1) && (movie->preRoll[i] != 0)) {
          return 1;
        }
        
        /* if preRoll == -1 we read the "preroll" AccessUnit from file */
        if (movie->preRoll[i] == -1) {
          
          u64 ext_dts, duration;
          unsigned int ext_cts, sampleDescrIdx;
          ISOHandle    tmp_mp4SampleH;
          unsigned int tmp_mp4SampleH_size;
          
          err = ISONewHandle( 0, &tmp_mp4SampleH ); if (err) break;
          
          err = MP4GetIndMediaSample(
            movie->media[i], 
            sampleNumber + movie->preRoll[i], 
            tmp_mp4SampleH, 
            &tmp_mp4SampleH_size, 
            &ext_dts, 
            &ext_cts, 
            &duration, 
            &sampleFlags, 
            &sampleDescrIdx
            );
          
          if(tmp_mp4SampleH_size <= (1<<16)){
            memcpy(sampleBuf[i], *tmp_mp4SampleH, tmp_mp4SampleH_size);
            sampleBufLen[i] = tmp_mp4SampleH_size;
          } else {
            return 1;
          }
          
          /* no output samples for this frame in case of preRoll */
          /* is this duration in movie/mediatime; if so, do we need to convert to samples?? */
          /* asume track last track as sls and responsible for sample output */
          *outSamples = -duration; 
          
          ISODisposeHandle(tmp_mp4SampleH);
        }
        
        ISODisposeHandle(description);
      }
    }
    
    if(movie->preRoll[i] == 0){
      if(movie->sampleBufH_size[i] <= (1<<16)){
        memcpy(sampleBuf[i], *movie->sampleBufH[i], movie->sampleBufH_size[i]);
        sampleBufLen[i] = movie->sampleBufH_size[i];
        *outSamples = movie->outSamples[i];
      } else {
        return 1;
      }
    }
  } /* for layer */
bail:	
  return err;
}
#endif


MP4Err MP4CloseMovie(MP4MovieStruct movie, unsigned int totalLayers)
{
	unsigned int i;
	MP4Err err=0;

	if(err = MP4DisposeMovie( movie.moov )) goto bail;
	for (i=0; i<totalLayers; i++)
	{
		if(err = MP4DisposeHandle( movie.sampleBufH[i] )) goto bail;
		if(err = MP4DisposeHandle( movie.decoderSpecificInfoH[i] )) goto bail;
	}
bail:
	return err;
}



/* encoder-related functions */
MP4Err MP4CreateNewMovie(MP4MovieStruct* movieOut, unsigned int corepresent, 
                         unsigned int samplingfreq, unsigned int AOT, unsigned int numchannels, 
                         unsigned int numbits, unsigned int OSF, unsigned int numlayers) 
{
	unsigned int i;
	MP4Err err=0;
#ifdef SAMPLEGROUP_ELST	
        unsigned int maxSamplingFrequency = 0;
#endif
	if(err = MP4NewMovie( &(movieOut->moov), 1, 0xff, 0xff, 0x01, 0xff, 0xff)) goto bail;
		
	for(i=0;i<numlayers;i++)
	{
		unsigned int outObjectType=0x40, outStreamType=0x05, outBufferSize=0, outMaxBitrate=0, 
			outAvgBitrate=0, streamPriority;

#ifdef USE_OLD_MP4A_TRACKSIGNALLING
		/* The core elementary stream (BSAC, AAC-lc or AAC-scal) and the enhancement elementary stream are marked 'mp4a' */
		if(err = MP4NewMovieTrack( movieOut->moov, MP4NewTrackIsAudio, &(movieOut->trak[i]) ))goto bail;
#else
		/* The core elementary stream (BSAC, AAC-lc or AAC-scal) is marked with 'mp4a', but all enhancement elementary streams are marked 'm4ae' (w10501) */
		if(err = MP4NewMovieTrack( movieOut->moov, MP4NewTrackIsAudio | ( i>0 ? MP4NewTrackIsEnhancement : 0), &(movieOut->trak[i]) ))goto bail;
#endif
		if(err = MP4AddTrackToMovieIOD( movieOut->trak[i] ))goto bail;
#ifndef SAMPLEGROUP_ELST
                if(err = MP4NewTrackMedia( movieOut->trak[i], &(movieOut->media[i]), MP4AudioHandlerType, samplingfreq, NULL ))goto bail;
#else
                if(corepresent&&(i==0)) {
		  if(err = MP4NewTrackMedia( movieOut->trak[i], &(movieOut->media[i]), MP4AudioHandlerType, samplingfreq, NULL ))goto bail;
		} else {
                   if(err = MP4NewTrackMedia( movieOut->trak[i], &(movieOut->media[i]), MP4AudioHandlerType, samplingfreq*OSF, NULL ))goto bail;
		}
#endif			
		if(err = MP4BeginMediaEdits( movieOut->media[i] ))goto bail;
		if(err = MP4SetMediaLanguage( movieOut->media[i], "und" ))goto bail;
	
		if(err = MP4NewHandle( 0, &(movieOut->sampleEntryH[i]) ))goto bail;
		if(err = MP4NewHandle( 0, &(movieOut->decoderSpecificInfoH[i]) ))goto bail;

#ifdef SAMPLEGROUP_ELST
		/* block for adding sample groups */
		{
      unsigned char* SGD_buff;
			unsigned int roll = ('r' << 24) | ('o' << 16) | ('l' << 8) | ('l' << 0);
			MP4Handle sampleGroupDescriptionH = NULL;
			err = MP4NewHandle(2, &sampleGroupDescriptionH);
			
      SGD_buff = (unsigned char *) *sampleGroupDescriptionH;
			SGD_buff[0] = 0xff; SGD_buff[1] = 0xff;

			ISOAddGroupDescription(movieOut->media[i], roll, sampleGroupDescriptionH, &movieOut->sampleDescriptionIndex[i]);
						
			err = MP4DisposeHandle(sampleGroupDescriptionH);
		}
#endif
		
		if(corepresent&&(i==0))
		{	
			char* ptrH;
	
			if(err = MP4NewHandle(9, &(movieOut->decoderSpecificInfoH[i]))) goto bail;
			ptrH = *movieOut->decoderSpecificInfoH[i];
#ifdef BSAC

#ifndef SAMPLEGROUP_ELST
			if(codec_type)
			{
                                /* AOT for BSAC */
				setGAAudioSpecificConfig(&(movieOut->decoderSpecificInfoH[i]), 22, 
					samplingfreq, numchannels, numbits, corepresent, OSF*1024, NULL);
			}
			else
			{
                                /* AOT for AAC LC : need to be added for other AOT such as AAC scalable */
				setGAAudioSpecificConfig(&(movieOut->decoderSpecificInfoH[i]), 2, 
					samplingfreq, numchannels, numbits, corepresent, OSF*1024, NULL);
                             
			}
#else
			if(codec_type)
			{
				/* AOT for BSAC */
				setGAAudioSpecificConfig(&(movieOut->decoderSpecificInfoH[i]), 22, 
					samplingfreq, numchannels, numbits, corepresent, 1024, NULL);
			}
			else
			{
                                /* AOT for AAC LC : need to be added for other AOT such as AAC scalable */
				setGAAudioSpecificConfig(&(movieOut->decoderSpecificInfoH[i]), 2, 
					samplingfreq, numchannels, numbits, corepresent, 1024, NULL);
			}

                        movieOut->samplesPerFrame[i] = 1024;

		        if(samplingfreq > maxSamplingFrequency){
			  maxSamplingFrequency = samplingfreq;
			}
#endif

#else
			/* config TAGS */
			ptrH[0] = 0x05; ptrH[1] = ptrH[2] = ptrH[3] = 0x80; ptrH[4] = 0x04;
			/* memcpy(ptrH+5,(char*) *(movie->decoderSpecificInfoH[i]), 4); */
#endif
			if(err = MP4NewSampleDescription(  movieOut->trak[i], movieOut->sampleEntryH[i]
                                                           ,1		  /*desc index */
                                                           ,outObjectType /* 0x40  //MPEG4 audio        */
                                                           ,outStreamType /* 0x05  //audio stream type  */
                                                           ,outBufferSize /*	   //decoderBufferSize  */
                                                           ,outMaxBitrate /* 0	   //max bitrate        */
                                                           ,outAvgBitrate /* 0	   //avg. bitrate       */
                                                           ,movieOut->decoderSpecificInfoH[i] )) goto bail;
	

		}
		else
		{
			unsigned int LLEmainstream = ((i-corepresent)==0)?1:0;

			setSLSAudioSpecificConfig(&(movieOut->decoderSpecificInfoH[i]), AOT, 
				samplingfreq*OSF, numchannels, numbits, corepresent, OSF*1024, LLEmainstream);

#ifdef SAMPLEGROUP_ELST
			movieOut->samplesPerFrame[i] = OSF*1024;

			if(samplingfreq*OSF > maxSamplingFrequency){
					maxSamplingFrequency = samplingfreq*OSF;
				}
#endif

			if(err = MP4NewSampleDescription(  movieOut->trak[i], movieOut->sampleEntryH[i]
                                                           ,1             /*desc index                       */
                                                           ,outObjectType /* 0x40	//MPEG4 audio        */
                                                           ,outStreamType /* 0x05	//audio stream type  */
                                                           ,outBufferSize /*            //decoderBufferSize  */
                                                           ,outMaxBitrate /* 0		//max bitrate        */
                                                           ,outAvgBitrate /* 0		//avg. bitrate       */
                                                           ,movieOut->decoderSpecificInfoH[i] ))goto bail;


                        {
                          if(i > 0){
			    int dummy;

			    /* track dependency */
			    MP4AddTrackReference( movieOut->trak[i],
						  movieOut->trak[i-1],
						  MP4StreamDependencyReferenceType,
						  &dummy
						  );
			  }
                        }

		}
	}

#ifdef SAMPLEGROUP_ELST
	if(maxSamplingFrequency > 0){
	  ISOSetMovieTimeScale(movieOut->moov, maxSamplingFrequency);
	}
#endif	

	for (i=0; i<numlayers; i++)
          if(err = MP4NewHandle( 1024*36, &(movieOut->sampleBufH[i]) )) goto bail; /* I2R (27 Jun 07) */
	
		

bail:
	return err;
}
#ifdef BSAC
MP4Err setGAAudioSpecificConfig(MP4Handle *decoderSpecificConfigH, unsigned int audioObjectType, 
		unsigned int samplingFreq, unsigned int channelConfig, unsigned int pcmWordLength, 
		unsigned int corePresent, unsigned int frameLength, unsigned int lle_main_stream)
{
	char* ptrH,*plen;
	unsigned int sampling_idx = 0xf;
        unsigned int pcm_idx, frame_idx, i, reserved_bit=0;
	unsigned int temp, *configPtr,n;
	unsigned char config[4];
	MP4Err err =0;	


	if(err = MP4NewHandle(256, decoderSpecificConfigH)) goto bail; /* MyExit("DSI handle err!"); */

	ptrH = **decoderSpecificConfigH;
	ptrH[0] = 0x05; ptrH[1] = ptrH[2] = ptrH[3] = 0x80; ptrH[4] = 0x08;  /* set max length to 8, e.g. for 192khz */

	/*GA Config*/
	if ((audioObjectType == 2) || (audioObjectType == 6) || (audioObjectType == 22))
	{	
		for ( i=0; i<13; i++)
			if( sampling_rate[i]== samplingFreq){ sampling_idx = i; break; }

		for ( i=0; i<8; i++)
			if( pcm_word_length[i]== pcmWordLength ){ pcm_idx = i; break; }

		for ( i=0; i<8; i++)
			if( frame_length[i]== frameLength ){ frame_idx = i; break; }

		if( sampling_idx != 0xf )
		{

                  /*Audio Specific Config*/
                  temp =	((audioObjectType &0x1F)<<27) | ((sampling_idx &0x0F)<<23) | ((channelConfig &0x0F)<<19);

                  configPtr = &temp;

                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
                  
                  memcpy(ptrH+5, (char*)config, 4);
                }
                else
                {
                  /* non-standard sample rates, e.g. 192khz */

                  temp = ((audioObjectType &0x1F)<<27) | ((sampling_idx &0x0F)<<23) | ((samplingFreq>>1) &0x7FFFFF);
                  
                  configPtr = &temp;
                  
                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
                  
                  memcpy(ptrH+5, (char*)config, 4);
                  
                  temp = ((samplingFreq &0x1)<<31) | ((channelConfig &0x0F)<<27);
                  
                  configPtr = &temp;
                  
                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
                  
                  memcpy(ptrH+9, (char*)config, 4);
                }

	} else {
          printf("unsupported audio object type %d!\n", audioObjectType);
          goto bail;
        }

bail:
	return err;
}	
#endif

MP4Err setSLSAudioSpecificConfig(MP4Handle* AAZ_decoderSpecificConfigH, unsigned int audioObjectType, 
		unsigned int samplingFreq, unsigned int channelConfig, unsigned int pcmWordLength, 
		unsigned int corePresent, unsigned int frameLength, unsigned int lle_main_stream)
{
	char* ptrH;
	unsigned int sampling_idx = 0xf;
        unsigned int pcm_idx, frame_idx, i, reserved_bit=0;
	unsigned int temp, *configPtr;
	unsigned char config[4];
	MP4Err err =0;	
	
	/* I2R: modified size to 13 (22 Jun 07) */
	if(err = MP4NewHandle(9+4, AAZ_decoderSpecificConfigH)) goto bail; /* MyExit("DSI handle err!"); */
	ptrH = **AAZ_decoderSpecificConfigH;
	
	/*config TAGS*/
	ptrH[0] = 0x05; ptrH[1] = ptrH[2] = ptrH[3] = 0x80; ptrH[4] = 0x08; /* I2R: modified 0x04 to 0x08 (22 Jun 07) */

	/*SLS Config*/
	if ( (audioObjectType == 37) || (audioObjectType == 38) )
	{
		unsigned int audioObjectTypeExt;
		
		audioObjectTypeExt = audioObjectType - 32;
		audioObjectType = 31;
        
		if((audioObjectTypeExt != 5) && (audioObjectTypeExt != 6))
                  goto bail; /*myExit("non-SLS audio object type!");*/
	
		/* I2R: changed from 12 to 13 to include 192kHz (28 May 07) */
		for ( i=0; i<13; i++)
			if( sampling_rate[i]== samplingFreq){ sampling_idx = i; break; }

		for ( i=0; i<8; i++)
			if( pcm_word_length[i]== pcmWordLength ){ pcm_idx = i; break; }

		for ( i=0; i<8; i++)
			if( frame_length[i]== frameLength ){ frame_idx = i; break; }

		/* I2R: modified to correctly include 192kHz case (21 Jun 07) */
		if( sampling_idx != 0xf )
		{
			
                  /*Audio Specific Config*/
                  temp = ((audioObjectType &0x1F)<<27) | ((audioObjectTypeExt &0x3F)<<21) |
                    ((sampling_idx &0x0F)<<17) | ((channelConfig &0x0F)<<13) |
                    
                    /*SLS Specific Config*/
                    ((pcm_idx & 0x07)<<10)		| (( corePresent &0x01)<<9)		|
                    ((lle_main_stream & 0x01)<<8)	| ((reserved_bit & 0x01)<<7)		| ((frame_idx&0x07)<<4) ;
	
                  configPtr = &temp;
	
                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
	
                  memcpy(ptrH+5, (char*)config, 4);
		}
		else
		{
                  /* sampfreq e.g. 192kHz */
	
                  /* I2R: max sampling index changed to 15 (escape value) to include 192kHz (21 Jun 07) */
                  /* if( samplingFreq==192000) sampling_idx = 15; */  /* not necessary, also for e.g. 176khz */
                  
                  /*Audio Specific Config*/
                  temp =	((audioObjectType &0x1F)<<27)	| ((audioObjectTypeExt &0x3F)<<21)	|
                    ((sampling_idx &0x0F)<<17)		| ((samplingFreq>>7) &0x1FFFF)		;
                  
                  configPtr = &temp;
                  
                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
                  
                  memcpy(ptrH+5, (char*)config, 4);
                  
                  temp =	((samplingFreq &0x7F)<<25)		| ((channelConfig &0x0F)<<21)		|
                    
                    /*SLS Specific Config*/
                    ((pcm_idx & 0x07)<<18)			| (( corePresent &0x01)<<17)			|
                    ((lle_main_stream & 0x01)<<16)	| ((reserved_bit & 0x01)<<15)		| ((frame_idx&0x07)<<12) ;
                  
                  configPtr = &temp;
                  
                  config[0] = ((unsigned char*)configPtr)[3];
                  config[1] = ((unsigned char*)configPtr)[2];
                  config[2] = ((unsigned char*)configPtr)[1];
                  config[3] = ((unsigned char*)configPtr)[0];
                  
                  memcpy(ptrH+9, (char*)config, 4);
		}
	}
	else
		fprintf(stderr, "Non-SLS audio object type!\n");
bail:
	return err;
}	

MP4Err MP4SetNextAUintoMovie(MP4MovieStruct* movieOut, unsigned int *sampleBufLen, 
							 unsigned int duration, unsigned int numlayers)
{
	unsigned int i, unitSize, sampleFlags, cts, dts;
	unsigned int iAAZTotalBufLen[2]={0};
	unsigned int LLEbufsize[2], next=0, ch=0;
	
	MP4Handle sampleDurationH, sampleBufSizeH;
	MP4Err err=0;

	if (err = MP4NewHandle( sizeof(u32), &sampleDurationH ))goto bail;
	if (err = MP4NewHandle( sizeof(u32), &sampleBufSizeH )) goto bail;
#ifndef SAMPLEGROUP_ELST
        *((unsigned int*)*sampleDurationH)=duration;

#endif	
	for(i=0;i<numlayers;i++)
	{
#ifdef SAMPLEGROUP_ELST
		*((unsigned int*)*sampleDurationH)= movieOut->samplesPerFrame[i];
#endif
          /*  *((unsigned int*) *sampleBufSizeH) = sampleBufLen[i]; */
		if(err = MP4SetHandleSize (movieOut->sampleBufH[i], sampleBufLen[i]))goto bail;
		if(err = MP4GetHandleSize( movieOut->sampleBufH[i], (unsigned int*) *sampleBufSizeH )) goto bail;
		if(err = MP4AddMediaSamples( movieOut->media[i], movieOut->sampleBufH[i], 1, sampleDurationH,
			sampleBufSizeH, movieOut->sampleEntryH[i], NULL, NULL )) goto bail;

		if ( movieOut->sampleEntryH[i] ) 
		{
			if(err = MP4DisposeHandle( movieOut->sampleEntryH[i] )) goto bail;
			movieOut->sampleEntryH[i] = NULL; 
		}	
	}
	if(err = MP4DisposeHandle( sampleDurationH  )) goto bail;
	if(err = MP4DisposeHandle( sampleBufSizeH  )) goto bail;
bail:	
	return err;
 }

#ifndef SAMPLEGROUP_ELST
MP4Err MP4OutputMovie(MP4MovieStruct* movieOut, const char* outfilename,
                                         unsigned int numlayers)
#else
MP4Err MP4OutputMovie(MP4MovieStruct* movieOut, 
											const char*  outfilename,
											unsigned int numlayers, 
											const int    nDelay,
											const int    sampleRateOut,
			                const int    originalSampleCount,
			                const int    originalSamplingRate)
#endif
{
	unsigned int i;
	u64 mediaDuration;
	MP4Err err=0;

#ifdef SAMPLEGROUP_ELST	
	u32 editBegin = 0;     /* in movie ticks */
       u64 editDuration = 0;  /* in media ticks */

	u32 mediaTimeScale = 0;
	u32 movieTimeScale = 0;
	u32 sampleCount    = 0;

	unsigned int roll = ('r' << 24) | ('o' << 16) | ('l' << 8) | ('l' << 0);
	
  err = MP4GetMovieTimeScale( movieOut->moov, &movieTimeScale ); if (err) goto bail;
#endif
	
	for(i=0;i<numlayers;i++)
	{
		if(err = MP4EndMediaEdits( movieOut->media[i] )) goto bail;
		if(err = MP4GetMediaDuration( movieOut->media[i], &mediaDuration ))goto bail;
	    
#ifndef SAMPLEGROUP_ELST
                if(err = MP4InsertMediaIntoTrack( movieOut->trak[i], 0, 0, mediaDuration, 1 ))goto bail;
#else		
		editDuration = mediaDuration;
		
                err = MP4GetMediaTimeScale( movieOut->media[i], &mediaTimeScale ); if (err) goto bail;
		if(originalSampleCount <= -1 || nDelay <= -1) {
		  editBegin = 0;
			editDuration = mediaDuration;
		}
		else {
		  editBegin = (u32)((float)nDelay * ((float)mediaTimeScale / (float)sampleRateOut));
			editDuration = (u64)((float)originalSampleCount * ((float) mediaTimeScale  / (float)originalSamplingRate));
		}

		MP4GetMediaSampleCount(movieOut->media[i], &sampleCount);
				
		ISOMapSamplestoGroup(movieOut->media[i], roll, movieOut->sampleDescriptionIndex[i], 0, sampleCount);
                err = MP4InsertMediaIntoTrack( movieOut->trak[i], 0, editBegin, editDuration, 1 ); if (err) goto bail;
#endif		
		if(err = MP4SetTrackEnabled( movieOut->trak[i], 1))goto bail;
	}
	
	if(err = MP4WriteMovieToFile( movieOut->moov, outfilename ))goto bail;
	if(err = MP4DisposeMovie( movieOut->moov ))goto bail;
	
	for(i=0;i<numlayers;i++)
	{
		if(err = MP4DisposeHandle( movieOut->sampleBufH[i])) goto bail;
		if(err = MP4DisposeHandle( movieOut->decoderSpecificInfoH[i] )) goto bail;
	}
bail:
	return err;
}
